from setuptools import setup

setup(
    name="amora",
    version="0.1.5rc2",
    description="Amora Data Build Tool",
    url="http://github.com/mundipagg/amora-data-build-tool",
    author="TREX Data",
    author_email="diogo.martins@stone.com.br",
    license="MIT",
    packages=["."],
    entry_points={
        "console_scripts": ["amora=cli:main"],
        "pytest11": ["amora = amora.tests.pytest_plugin"],
    },
    install_requires=[
        "matplotlib~=3.4.2",
        "networkx~=2.6.3",
        "numpy~=1.21.1",
        "pandas~=1.3.0",
        "pytest~=6.2.5",
        "pytest-xdist[psutil]~=2.4.0",
        "sqlalchemy-bigquery~=1.2.0",
        "sqlmodel~=0.0.4",
        "typer[all]~=0.4.0",
        "Jinja2~=2.11.3",
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: SQL",
    ],
)



from typing import Iterable, Optional, Callable

import pytest
from sqlalchemy import func, and_, Table, text, literal_column, union_all
from sqlalchemy.orm import InstrumentedAttribute
from sqlmodel.sql.expression import SelectOfScalar

from amora.models import select, Compilable, AmoraModel
from amora.compilation import compile_statement
from amora.providers.bigquery import get_client

Column = InstrumentedAttribute
Columns = Iterable[Column]
Test = Callable[..., SelectOfScalar]


def _test(statement: Compilable) -> bool:
    sql_stmt = compile_statement(statement=statement)

    query_job = get_client().query(sql_stmt)
    result = query_job.result()

    if result.total_rows == 0:
        return True
    else:
        pytest.fail(
            f"{result.total_rows} rows failed the test assertion."
            f"\n==========="
            f"\nTest query:"
            f"\n==========="
            f"\n{sql_stmt}",
            pytrace=False,
        )


def that(
    column: Column,
    test: Test,
    **test_kwargs,
) -> bool:
    """
    >>> assert that(HeartRate.value, is_not_null)

    Executes the test, returning True if the test is successful and raising a pytest fail otherwise
    """
    return _test(statement=test(column, **test_kwargs))


def is_not_null(column: Column) -> Compilable:
    """
    >>> is_not_null(HeartRate.id)

    The `id` column in the `HeartRate` model should not contain `null` values

    Results in the following query:

    ```sql
        SELECT {{ column_name }}
        FROM {{ model }}
        WHERE {{ column_name }} IS NULL
    ```
    """
    return select(column).where(column == None)


def is_unique(column: Column) -> Compilable:
    """
    >>> is_unique(HeartRate.id)

    The `id` column in the `HeartRate` model should be unique

    ```sql
        SELECT {{ column_name }}
        FROM (
            SELECT {{ column_name }}
            FROM {{ model }}
            WHERE {{ column_name }} IS NOT NULL
            GROUP BY {{ column_name }}
            HAVING COUNT(*) > 1
        ) validation_errors
    ```
    """
    return select(column).group_by(column).having(func.count(column) > 1)


def has_accepted_values(column: Column, values: Iterable) -> Compilable:
    """
    >>> has_accepted_values(HeartRate.source, values=["iPhone", "Mi Band"])

    The `source` column in the `HeartRate` model should be one of
    'iPhone' or 'MiBand'

    ```sql
        SELECT {{ column_name }}
        FROM {{ model }}
        WHERE {{ column_name }} NOT IN {{ values }}
    ```
    """
    return select(column).where(~column.in_(values))


def relationship(
    from_: Column,
    to: Column,
    from_condition=and_(True),
    to_condition=and_(True),
) -> bool:
    """
    >>> relationship(HeartRate.id, to=Health.id)

    Each `id` in the `HeartRate` model exists as an `id` in the `Health`
    table (also known as referential integrity)

    This test validates the referential integrity between two relations
    with a predicate (`from_condition` and `to_condition`) to filter out
    some rows from the test. This is useful to exclude records such as
    test entities, rows created in the last X minutes/hours to account
    for temporary gaps due to data ingestion limitations, etc.

    ```sql
        WITH left_table AS (
          SELECT
            {{from_column_name}} AS id
          FROM {{from_table}}
          WHERE
            {{from_column_name}} IS NOT NULL
            AND {{from_condition}}
        ),
        right_table AS (
          SELECT
            {{to_column_name}} AS id
          FROM {{to_table}}
          WHERE
            {{to_column_name}} IS NOT NULL
            AND {{to_condition}}
        ),
        exceptions as (
          SELECT
            left_table.id AS {{from_column_name}}}
          FROM
            left_table
          LEFT JOIN
            right_table
            ON left_table.id = right_table.id
          WHERE
            right_table.id IS NULL
        )

        SELECT * FROM exceptions
    ```

    """
    left_table = (
        select(from_.label("id"))
        .where(from_ != None)
        .where(from_condition)
        .cte("left_table")
    )
    right_table = (
        select(to.label("id"))
        .where(to != None)
        .where(to_condition)
        .cte("right_table")
    )

    exceptions = (
        select([left_table.c["id"].label(from_.key)])
        .select_from(
            left_table.join(
                right_table,
                onclause=left_table.c["id"] == right_table.c["id"],
                isouter=True,
            )
        )
        .where(right_table.c["id"] == None)
    )

    return _test(statement=exceptions)


def is_non_negative(column: Column) -> Compilable:
    """
    >>> is_non_negative(HeartRate.value)
    True

    Each not null `value` in `HeartRate` model is >= 0

    ```sql
        SELECT {{ column_name }}
        FROM {{ model }}
        WHERE {{ column_name }} < 0
    ```
    """
    return select(column).where(column < 0)


def expression_is_true(expression, condition=and_(True)) -> bool:
    """
    >>> expression_is_true(StepsAgg._sum > StepsAgg._avg, condition=StepsAgg.year == 2021)

    Asserts that a expression is TRUE for all records.
    This is useful when checking integrity across columns, for example,
    that a total is equal to the sum of its parts, or that at least one column is true.

    Optionally assert `expression` only for rows where `condition` is met.
    ```
    """
    return _test(statement=select(["*"]).where(condition).where(~expression))


def equality(
    model_a: AmoraModel,
    model_b: AmoraModel,
    compare_columns: Optional[Iterable[Column]] = None,
) -> bool:
    """
    This schema test asserts the equality of two models. Optionally specify a subset of columns to compare.

    """

    raise NotImplementedError

    def comparable_columns(model: AmoraModel) -> Iterable[Column]:
        if not compare_columns:
            return model
        return [getattr(model, column_name) for column_name in compare_columns]

    a = select(comparable_columns(model_a)).cte("a")
    b = select(comparable_columns(model_b)).cte("b")

    # fixme: google.api_core.exceptions.BadRequest: 400 EXCEPT must be followed by ALL, DISTINCT, or "(" at [34:4]
    a_minus_b = select(a).except_(select(b))
    b_minus_a = select(b).except_(select(a))

    diff_union = union_all(a_minus_b, b_minus_a)

    return _test(statement=diff_union)


def has_at_least_one_not_null_value(column: Column) -> Compilable:
    """
    Asserts if column has at least one value.

    ```sql

    SELECT
        count({{ column_name }}) as filler_column
    FROM
        {{ model }}
    HAVING
        count({{ column_name }}) = 0
    ```
    """
    return select(func.count(column)).having(func.count(column) == 0)


def are_unique_together(columns: Iterable[Column]) -> Compilable:
    """
    This test confirms that the combination of columns is unique.
    For example, the combination of month and product is unique,
    however neither column is unique in isolation.

    """
    return select(columns).group_by(*columns).having(func.count() > 1)



from dataclasses import dataclass, field
from datetime import datetime, date, time
from typing import Optional, List, Union

from google.cloud.bigquery import (
    Client,
    QueryJobConfig,
    SchemaField,
    Table,
    TableReference,
)

from amora.compilation import compile_statement
from amora.models import AmoraModel

Schema = List[SchemaField]
BQTable = Union[Table, TableReference, str]

# todo: cobrir todos os tipos
BIGQUERY_TYPES_TO_PYTHON_TYPES = {
    "ARRAY": list,
    "BIGNUMERIC": int,
    "BOOL": bool,
    "BOOLEAN": bool,
    "BYTES": bytes,
    "DATE": date,
    "DATETIME": datetime,
    "FLOAT64": float,
    "FLOAT": float,
    "GEOGRAPHY": str,
    "INT64": int,
    "INTEGER": int,
    "JSON": dict,
    "STRING": str,
    "TIME": time,
    "TIMESTAMP": datetime,
}


@dataclass
class DryRunResult:
    total_bytes: int
    model: AmoraModel
    schema: Schema
    query: Optional[str] = None
    referenced_tables: List[str] = field(default_factory=list)

    @property
    def estimated_cost(self):
        raise NotImplementedError


_client = None


def get_client() -> Client:
    global _client
    if _client is None:
        _client = Client()
    return _client


def get_fully_qualified_id(model: AmoraModel) -> str:
    return f"{model.metadata.schema}.{model.__tablename__}"


def get_schema(table_id: str) -> Schema:
    client = get_client()
    table = client.get_table(table_id)
    return table.schema


def dry_run(model: AmoraModel) -> Optional[DryRunResult]:
    """
    >>> dry_run(HeartRate)
    DryRunResult(
        total_bytes_processed=170181834,
        query="SELECT\n  `health`.`creationDate`,\n  `health`.`device`,\n  `health`.`endDate`,\n  `health`.`id`,\n  `health`.`sourceName`,\n  `health`.`startDate`,\n  `health`.`unit`,\n  `health`.`value`\nFROM `diogo`.`health`\nWHERE `health`.`type` = 'HeartRate'",
        model=HeartRate,
        referenced_tables=["stg-tau-rex.diogo.health"],
        schema=[
            SchemaField("creationDate", "TIMESTAMP", "NULLABLE", None, (), None),
            SchemaField("device", "STRING", "NULLABLE", None, (), None),
            SchemaField("endDate", "TIMESTAMP", "NULLABLE", None, (), None),
            SchemaField("id", "INTEGER", "NULLABLE", None, (), None),
            SchemaField("sourceName", "STRING", "NULLABLE", None, (), None),
            SchemaField("startDate", "TIMESTAMP", "NULLABLE", None, (), None),
            SchemaField("unit", "STRING", "NULLABLE", None, (), None),
            SchemaField("value", "FLOAT", "NULLABLE", None, (), None),
        ],
    )

    You can use the estimate returned by the dry run to calculate query
    costs in the pricing calculator. Also useful to verify user permissions
    and query validity. You are not charged for performing the dry run.

    Read more: https://cloud.google.com/bigquery/docs/dry-run-queries
    """
    client = get_client()
    source = model.source()
    if source is None:
        table = client.get_table(get_fully_qualified_id(model))

        if table.table_type == "VIEW":
            query_job = client.query(
                query=table.view_query,
                job_config=QueryJobConfig(dry_run=True, use_query_cache=False),
            )

            return DryRunResult(
                model=model,
                query=table.view_query,
                referenced_tables=[
                    ".".join(table.to_api_repr().values())
                    for table in query_job.referenced_tables
                ],
                schema=query_job.schema,
                total_bytes=query_job.total_bytes_processed,
            )
        else:
            return DryRunResult(
                total_bytes=table.num_bytes, schema=table.schema, model=model
            )

    query = compile_statement(source)

    query_job = client.query(
        query=query,
        job_config=QueryJobConfig(dry_run=True, use_query_cache=False),
    )
    tables = [table.to_api_repr() for table in query_job.referenced_tables]

    return DryRunResult(
        total_bytes=query_job.total_bytes_processed,
        referenced_tables=[".".join(table.values()) for table in tables],
        query=query,
        model=model,
        schema=query_job.schema,
    )



import inspect
import os
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import Optional, Protocol, runtime_checkable, Iterable

import sqlparse
from amora.config import settings
from amora.models import (
    AmoraModel,
    Compilable,
    list_target_files,
)
import sqlalchemy
from sqlalchemy_bigquery import BigQueryDialect
from sqlalchemy_bigquery.base import BigQueryCompiler, unnest


class fixed_unnest(sqlalchemy.sql.roles.InElementRole, unnest):
    def __init__(self, *args, **kwargs):
        self.name = "unnest"
        super().__init__(*args, **kwargs)


class AmoraBigQueryCompiler(BigQueryCompiler):
    def visit_getitem_binary(self, binary, operator_, **kw):
        left = self.process(binary.left, **kw)
        right = self.process(binary.right, **kw)

        try:
            # Only integer values should be wrapped in OFFSET
            return f"{left}[OFFSET({int(right)})]"
        except ValueError:
            return f"{left}[{right}]"


dialect = BigQueryDialect()
dialect.statement_compiler = AmoraBigQueryCompiler


@runtime_checkable
class CompilableProtocol(Protocol):
    def source(self) -> Compilable:
        ...


def compile_statement(statement: Compilable) -> str:
    raw_sql = str(
        statement.compile(
            dialect=dialect, compile_kwargs={"literal_binds": True}
        )
    )
    formatted_sql = sqlparse.format(raw_sql, reindent=True, indent_columns=True)
    return formatted_sql


def model_path_for_target_path(path: Path) -> Path:
    return Path(
        str(path)
        .replace(settings.TARGET_PATH, settings.MODELS_PATH)
        .replace(".sql", ".py"),
    )


def target_path_for_model_path(path: Path) -> Path:
    return Path(
        str(path)
        .replace(settings.MODELS_PATH, settings.TARGET_PATH)
        .replace(".py", ".sql")
    )


def amora_model_for_path(path: Path) -> AmoraModel:
    spec = spec_from_file_location(path.stem, path)
    if spec is None:
        raise ValueError(f"Invalid path `{path}`. Not a valid Python file.")

    module = module_from_spec(spec)
    # if module is None:
    #     raise ValueError(f"Invalid path `{path}`")

    if spec.loader is None:
        raise ValueError(f"Invalid path `{path}`. Unable to load module.")

    try:
        spec.loader.exec_module(module)  # type: ignore
    except ImportError as e:
        raise ValueError(
            f"Invalid path `{path}`. Unable to load module."
        ) from e

    compilables = inspect.getmembers(
        module,
        lambda x: isinstance(x, CompilableProtocol)
        and inspect.isclass(x)
        and issubclass(x, AmoraModel),  # type: ignore
    )

    for _name, class_ in compilables:
        try:
            # fixme: Quando carregamos o código em inspect, não existe um arquivo associado,
            #  ou seja, ao iterar sobre as classes de um arquivo, a classe que retornar um TypeError,
            #  é uma classe definida no arquivo
            class_.model_file_path()
        except TypeError:
            return class_

    raise ValueError(f"Invalid path `{path}`")


def amora_model_for_target_path(path: Path) -> AmoraModel:
    model_path = model_path_for_target_path(path)
    return amora_model_for_path(model_path)


def clean_compiled_files():
    for sql_file in list_target_files():
        os.remove(sql_file)
